from __future__ import annotations

import contextlib
import datetime
import os
import pathlib
import tempfile
from pathlib import Path

import pyarrow as pa
import pyarrow.parquet as papq
import pytest

import daft
from daft.datatype import DataType, TimeUnit
from daft.exceptions import DaftCoreException
from daft.logical.schema import Schema
from daft.recordbatch import (
    MicroPartition,
    read_parquet_into_pyarrow,
    read_parquet_into_pyarrow_bulk,
    recordbatch_io,
)
from daft.runners.partitioning import TableParseParquetOptions, TableReadOptions

PYARROW_GE_11_0_0 = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) >= (11, 0, 0)
PYARROW_GE_13_0_0 = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) >= (13, 0, 0)


def test_read_input(tmpdir):
    tmpdir = pathlib.Path(tmpdir)
    data = pa.Table.from_pydict({"foo": [1, 2, 3]})
    with open(tmpdir / "file.parquet", "wb") as f:
        papq.write_table(data, f)

    schema = Schema._from_field_name_and_types([("foo", DataType.int64())])

    # Test pathlib, str and IO
    assert recordbatch_io.read_parquet(tmpdir / "file.parquet", schema=schema).to_arrow() == data
    assert recordbatch_io.read_parquet(str(tmpdir / "file.parquet"), schema=schema).to_arrow() == data


@contextlib.contextmanager
def _parquet_write_helper(data: pa.Table, row_group_size: int | None = None, papq_write_table_kwargs: dict = {}):
    with tempfile.TemporaryDirectory() as directory_name:
        file = os.path.join(directory_name, "tempfile")
        papq.write_table(data, file, row_group_size=row_group_size, **papq_write_table_kwargs)
        yield file


@pytest.mark.parametrize(
    ["data", "expected_dtype"],
    [
        ("1", DataType.string()),
        ("foo", DataType.string()),
        ("1.5", DataType.string()),
        ("True", DataType.string()),
        ("None", DataType.string()),
        (1, DataType.int64()),
        (1.5, DataType.float64()),
        (True, DataType.bool()),
        (None, DataType.null()),
        ({"foo": 1}, DataType.struct({"foo": DataType.int64()})),
        ([1, None, 2], DataType.list(DataType.int64())),
    ],
)
def test_parquet_infer_schema(data, expected_dtype):
    with _parquet_write_helper(
        pa.Table.from_pydict(
            {
                "id": [1, 2, 3],
                "data": [data, data, None],
            }
        )
    ) as f:
        schema = Schema.from_parquet(f)
        assert schema == Schema._from_field_name_and_types([("id", DataType.int64()), ("data", expected_dtype)])


def test_parquet_read_empty():
    with _parquet_write_helper(pa.Table.from_pydict({"foo": pa.array([], type=pa.int64())})) as f:
        schema = Schema._from_field_name_and_types([("foo", DataType.int64())])
        expected = MicroPartition.from_pydict({"foo": pa.array([], type=pa.int64())})
        table = recordbatch_io.read_parquet(f, schema)
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


@pytest.mark.parametrize(
    ["data", "expected_data_series"],
    [
        [1, daft.Series.from_pylist([1, 1, None])],
        ["foo", daft.Series.from_pylist(["foo", "foo", None])],
        [1.5, daft.Series.from_pylist([1.5, 1.5, None])],
        (True, daft.Series.from_pylist([True, True, None])),
        (
            datetime.date(year=2021, month=1, day=1),
            daft.Series.from_pylist(
                [datetime.date(year=2021, month=1, day=1), datetime.date(year=2021, month=1, day=1), None]
            ),
        ),
        ([[], [1, 2], None], daft.Series.from_pylist([[[], [1, 2], None], [[], [1, 2], None], None])),
        ({"foo": 1}, daft.Series.from_pylist([{"foo": 1}, {"foo": 1}, None])),
    ],
)
def test_parquet_read_data(data, expected_data_series):
    with _parquet_write_helper(
        pa.Table.from_pydict(
            {
                "id": [1, 2, 3],
                "data": [data, data, None],
            }
        )
    ) as f:
        schema = Schema._from_field_name_and_types(
            [("id", DataType.int64()), ("data", expected_data_series.datatype())]
        )
        expected = MicroPartition.from_pydict(
            {
                "id": [1, 2, 3],
                "data": expected_data_series,
            }
        )
        table = recordbatch_io.read_parquet(f, schema)
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


@pytest.mark.parametrize("row_group_size", [None, 1, 3])
def test_parquet_read_data_limit_rows(row_group_size):
    with _parquet_write_helper(
        pa.Table.from_pydict(
            {
                "id": [1, 2, 3],
                "data": [1, 2, None],
            }
        ),
        row_group_size=row_group_size,
    ) as f:
        schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.int64())])
        expected = MicroPartition.from_pydict(
            {
                "id": [1, 2],
                "data": [1, 2],
            }
        )
        table = recordbatch_io.read_parquet(f, schema, read_options=TableReadOptions(num_rows=2))
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


def test_parquet_read_data_multi_row_groups():
    path = "tests/assets/parquet-data/mvp.parquet"
    table = MicroPartition.read_parquet(path)
    expected = MicroPartition.from_arrow(papq.read_table(path))
    assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


def test_parquet_read_data_select_columns():
    with _parquet_write_helper(
        pa.Table.from_pydict(
            {
                "id": [1, 2, 3],
                "data": [1, 2, None],
            }
        )
    ) as f:
        schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.int64())])
        expected = MicroPartition.from_pydict(
            {
                "data": [1, 2, None],
            }
        )
        table = recordbatch_io.read_parquet(f, schema, read_options=TableReadOptions(column_names=["data"]))
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


###
# Test Parquet Int96 timestamps
###


@pytest.mark.parametrize("use_deprecated_int96_timestamps", [True, False])
def test_parquet_read_int96_timestamps(use_deprecated_int96_timestamps):
    data = {
        "timestamp_ms": pa.array([1, 2, 3], pa.timestamp("ms")),
        "timestamp_us": pa.array([1, 2, 3], pa.timestamp("us")),
    }
    schema = [
        ("timestamp_ms", DataType.timestamp(TimeUnit.ms())),
        ("timestamp_us", DataType.timestamp(TimeUnit.us())),
    ]
    # int64 timestamps cannot support nanosecond resolutions
    if use_deprecated_int96_timestamps:
        data["timestamp_ns"] = pa.array([1, 2, 3], pa.timestamp("ns"))
        schema.append(("timestamp_ns", DataType.timestamp(TimeUnit.ns())))

    papq_write_table_kwargs = {
        "use_deprecated_int96_timestamps": use_deprecated_int96_timestamps,
        "coerce_timestamps": "us" if not use_deprecated_int96_timestamps else None,
    }
    if PYARROW_GE_11_0_0:
        papq_write_table_kwargs["store_schema"] = False

    with _parquet_write_helper(
        pa.Table.from_pydict(data),
        papq_write_table_kwargs=papq_write_table_kwargs,
    ) as f:
        schema = Schema._from_field_name_and_types(schema)
        expected = MicroPartition.from_pydict(data)
        table = recordbatch_io.read_parquet(
            f,
            schema,
            read_options=TableReadOptions(column_names=schema.column_names()),
        )
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


@pytest.mark.parametrize("coerce_to", [TimeUnit.ms(), TimeUnit.us()])
def test_parquet_read_int96_timestamps_overflow(coerce_to):
    # NOTE: datetime.datetime(3000, 1, 1) and datetime.datetime(1000, 1, 1) cannot be represented by our timestamp64(nanosecond)
    # type. However they can be written to Parquet's INT96 type. Here we test that a round-trip is possible if provided with
    # the appropriate flags.
    data = {
        "timestamp": pa.array(
            [datetime.datetime(1000, 1, 1), datetime.datetime(2000, 1, 1), datetime.datetime(3000, 1, 1)],
            pa.timestamp(str(coerce_to)),
        ),
    }
    schema = [
        ("timestamp", DataType.timestamp(coerce_to)),
    ]

    papq_write_table_kwargs = {
        "use_deprecated_int96_timestamps": True,
    }
    if PYARROW_GE_11_0_0:
        papq_write_table_kwargs["store_schema"] = False

    with _parquet_write_helper(
        pa.Table.from_pydict(data),
        papq_write_table_kwargs=papq_write_table_kwargs,
    ) as f:
        schema = Schema._from_field_name_and_types(schema)
        expected = MicroPartition.from_pydict(data)
        table = recordbatch_io.read_parquet(
            f,
            schema,
            read_options=TableReadOptions(column_names=schema.column_names()),
            parquet_options=TableParseParquetOptions(coerce_int96_timestamp_unit=coerce_to),
        )
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


@pytest.mark.parametrize("coerce_to", [TimeUnit.ms(), TimeUnit.us()])
@pytest.mark.parametrize("store_schema", [True, False])
def test_parquet_read_int96_timestamps_schema_inference(coerce_to, store_schema):
    dt = datetime.datetime(2000, 1, 1)
    ns_ts_array = pa.array(
        [dt, dt, dt],
        pa.timestamp("ns"),
    )
    data = {
        "timestamp": ns_ts_array,
        "nested_timestamp": pa.array([[dt], [dt], [dt]], type=pa.list_(pa.timestamp("ns"))),
        "struct_timestamp": pa.array([{"foo": dt} for _ in range(3)], type=pa.struct({"foo": pa.timestamp("ns")})),
        "struct_nested_timestamp": pa.array(
            [{"foo": [dt]} for _ in range(3)], type=pa.struct({"foo": pa.list_(pa.timestamp("ns"))})
        ),
    }
    schema = [
        ("timestamp", DataType.timestamp(coerce_to)),
        ("nested_timestamp", DataType.list(DataType.timestamp(coerce_to))),
        ("struct_timestamp", DataType.struct({"foo": DataType.timestamp(coerce_to)})),
        ("struct_nested_timestamp", DataType.struct({"foo": DataType.list(DataType.timestamp(coerce_to))})),
    ]
    expected = Schema._from_field_name_and_types(schema)

    papq_write_table_kwargs = {
        "use_deprecated_int96_timestamps": True,
    }
    if PYARROW_GE_11_0_0:
        papq_write_table_kwargs["store_schema"] = store_schema

    with _parquet_write_helper(
        pa.Table.from_pydict(data),
        papq_write_table_kwargs=papq_write_table_kwargs,
    ) as f:
        schema = Schema.from_parquet(f, coerce_int96_timestamp_unit=coerce_to)
        assert schema == expected, f"Expected:\n{expected}\n\nReceived:\n{schema}"


@pytest.mark.parametrize("n_bytes", [0, 1, 2, 7])
def test_read_too_small_parquet_file(tmpdir, n_bytes):
    tmpdir = pathlib.Path(tmpdir)
    file_path = tmpdir / "file.parquet"
    with open(file_path, "wb") as f:
        for _ in range(n_bytes):
            f.write(b"0")
    with pytest.raises(ValueError, match="smaller than the minimum size of 12 bytes"):
        MicroPartition.read_parquet(file_path.as_posix())


def test_read_empty_parquet_file_with_table(tmpdir):
    tmpdir = pathlib.Path(tmpdir)
    file_path = tmpdir / "file.parquet"
    tab = pa.table({"x": pa.array([], type=pa.int64())})
    papq.write_table(tab, file_path.as_posix())
    read_back = MicroPartition.read_parquet(file_path.as_posix()).to_arrow()
    assert tab == read_back


def test_read_empty_parquet_file_with_pyarrow(tmpdir):
    tmpdir = pathlib.Path(tmpdir)
    file_path = tmpdir / "file.parquet"
    tab = pa.table({"x": pa.array([], type=pa.int64())})
    papq.write_table(tab, file_path.as_posix())
    read_back = read_parquet_into_pyarrow(file_path.as_posix())
    assert tab == read_back


def test_read_parquet_file_missing_column_with_pyarrow(tmpdir):
    tmpdir = pathlib.Path(tmpdir)
    file_path = tmpdir / "file.parquet"
    tab = pa.table({"x": pa.array([1, 2, 3], type=pa.int64())})
    papq.write_table(tab, file_path.as_posix())
    read_back = read_parquet_into_pyarrow(file_path.as_posix(), columns=["MISSING"])
    assert tab.drop("x") == read_back  # same length, but no columns, as original table


def test_read_parquet_file_missing_column_partial_read_with_pyarrow(tmpdir):
    tmpdir = pathlib.Path(tmpdir)
    file_path = tmpdir / "file.parquet"
    tab = pa.table({"x": pa.array([1, 2, 3], type=pa.int64()), "y": pa.array([1, 2, 3], type=pa.int64())})
    papq.write_table(tab, file_path.as_posix())
    read_back = read_parquet_into_pyarrow(file_path.as_posix(), columns=["x", "MISSING"])
    assert tab.drop("y") == read_back  # only read "x"


def test_read_empty_parquet_file_with_pyarrow_bulk(tmpdir):
    tmpdir = pathlib.Path(tmpdir)
    file_path = tmpdir / "file.parquet"
    tab = pa.table({"x": pa.array([], type=pa.int64())})
    papq.write_table(tab, file_path.as_posix())
    read_back = read_parquet_into_pyarrow_bulk([file_path.as_posix()])
    assert len(read_back) == 1
    assert tab == read_back[0]


def test_read_parquet_file_missing_column_with_pyarrow_bulk(tmpdir):
    tmpdir = pathlib.Path(tmpdir)
    file_path = tmpdir / "file.parquet"
    tab = pa.table({"x": pa.array([1, 2, 3], type=pa.int64())})
    papq.write_table(tab, file_path.as_posix())
    read_back = read_parquet_into_pyarrow_bulk([file_path.as_posix()], columns=["MISSING"])
    assert len(read_back) == 1
    assert tab.drop("x") == read_back[0]  # same length, but no columns, as original table


def test_read_parquet_file_missing_column_partial_read_with_pyarrow_bulk(tmpdir):
    tmpdir = pathlib.Path(tmpdir)
    file_path = tmpdir / "file.parquet"
    tab = pa.table({"x": pa.array([1, 2, 3], type=pa.int64()), "y": pa.array([1, 2, 3], type=pa.int64())})
    papq.write_table(tab, file_path.as_posix())
    read_back = read_parquet_into_pyarrow_bulk([file_path.as_posix()], columns=["x", "MISSING"])
    assert len(read_back) == 1
    assert tab.drop("y") == read_back[0]  # only read "x"


@pytest.mark.parametrize(
    "parquet_path", [Path(__file__).parents[2] / "assets" / "parquet-data" / "invalid_utf8.parquet"]
)
def test_parquet_read_string_utf8_into_binary(parquet_path: Path):
    import pyarrow as pa

    assert parquet_path.exists()

    with pytest.raises(DaftCoreException, match="invalid utf-8 sequence"):
        read_parquet_into_pyarrow(path=parquet_path.as_posix())

    read_back = read_parquet_into_pyarrow(path=parquet_path.as_posix(), string_encoding="raw")
    schema = read_back.schema
    assert len(schema) == 1
    assert schema[0].name == "invalid_string"
    assert schema[0].type == pa.binary()
    assert read_back["invalid_string"][0].as_py() == b"\x80\x80\x80"



from __future__ import annotations

import numpy as np
import pyarrow as pa
import pytest

from daft.datatype import DataType
from daft.series import Series
from daft.utils import pyarrow_supports_fixed_shape_tensor
from tests.conftest import get_tests_daft_runner_name
from tests.series import ARROW_FLOAT_TYPES, ARROW_INT_TYPES, ARROW_STRING_TYPES

ARROW_VERSION = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())


@pytest.mark.parametrize("dtype", ARROW_INT_TYPES + ARROW_FLOAT_TYPES + ARROW_STRING_TYPES)
def test_series_filter(dtype) -> None:
    data = pa.array([1, 2, 3, None, 5, None])

    s = Series.from_arrow(data.cast(dtype))
    pymask = [False, True, None, True, False, False]
    mask = Series.from_pylist(pymask)

    result = s.filter(mask)

    assert s.datatype() == result.datatype()
    expected = [val for val, keep in zip(s.to_pylist(), pymask) if keep]
    assert result.to_pylist() == expected


def test_series_filter_on_null() -> None:
    s = Series.from_pylist([None, None, None, None, None, None])
    pymask = [False, True, None, True, False, False]
    mask = Series.from_pylist(pymask)

    result = s.filter(mask)

    assert s.datatype() == result.datatype()

    expected = [None, None]
    assert result.to_pylist() == expected


def test_series_filter_on_bool() -> None:
    s = Series.from_pylist([True, False, None, True, None, False])
    pymask = [False, True, True, None, False, False]
    mask = Series.from_pylist(pymask)

    result = s.filter(mask)

    assert s.datatype() == result.datatype()

    expected = [False, None]
    assert result.to_pylist() == expected


@pytest.mark.parametrize("type", [pa.binary(), pa.binary(1)])
def test_series_filter_on_binary(type) -> None:
    s = Series.from_arrow(pa.array([b"Y", b"N", None, b"Y", None, b"N"], type=type))
    pymask = [False, True, True, None, False, False]
    mask = Series.from_pylist(pymask)

    result = s.filter(mask)

    assert s.datatype() == result.datatype()

    expected = [b"N", None]
    assert result.to_pylist() == expected


@pytest.mark.parametrize("dtype", ARROW_INT_TYPES + ARROW_FLOAT_TYPES)
def test_series_filter_on_list_array(dtype) -> None:
    data = pa.array([[1], [2, None], None, [5, 6, 7], [8]], type=pa.list_(dtype))

    s = Series.from_arrow(data)
    pymask = [False, True, True, None, False]
    mask = Series.from_pylist(pymask)

    result = s.filter(mask)

    assert s.datatype() == result.datatype()
    expected = [val for val, keep in zip(s.to_pylist(), pymask) if keep]
    assert result.to_pylist() == expected


@pytest.mark.parametrize("dtype", ARROW_INT_TYPES + ARROW_FLOAT_TYPES)
def test_series_filter_on_fixed_size_list_array(dtype) -> None:
    data = pa.array([[1, 2], [3, None], None, [8, 9], [None, 11]], type=pa.list_(dtype, 2))

    s = Series.from_arrow(data)
    pymask = [False, True, True, None, False]
    mask = Series.from_pylist(pymask)

    result = s.filter(mask)

    assert s.datatype() == result.datatype()
    expected = [val for val, keep in zip(s.to_pylist(), pymask) if keep]
    assert result.to_pylist() == expected


def test_series_filter_on_struct_array() -> None:
    dtype = pa.struct({"a": pa.int64(), "b": pa.float64(), "c": pa.string()})
    data = pa.array(
        [{"a": 1, "b": 2}, {"b": 3, "c": "4"}, None, {"a": 5, "b": 6, "c": "7"}, {"a": 8, "b": None, "c": "10"}],
        type=dtype,
    )

    s = Series.from_arrow(data.cast(dtype))
    pymask = [False, True, True, None, False]
    mask = Series.from_pylist(pymask)

    result = s.filter(mask)

    assert s.datatype() == result.datatype()
    expected = [val for val, keep in zip(s.to_pylist(), pymask) if keep]
    assert result.to_pylist() == expected


@pytest.mark.skipif(
    get_tests_daft_runner_name() == "ray",
    reason="pyarrow extension types aren't supported on Ray clusters.",
)
def test_series_filter_on_extension_array(uuid_ext_type) -> None:
    arr = pa.array(f"{i}".encode() for i in range(5))
    data = pa.ExtensionArray.from_storage(uuid_ext_type, arr)

    s = Series.from_arrow(data)
    pymask = [False, True, True, None, False]
    mask = Series.from_pylist(pymask)

    result = s.filter(mask)

    assert s.datatype() == result.datatype()
    expected = [val for val, keep in zip(s.to_pylist(), pymask) if keep]
    assert result.to_pylist() == expected


@pytest.mark.skipif(
    not pyarrow_supports_fixed_shape_tensor(),
    reason=f"Arrow version {ARROW_VERSION} doesn't support the canonical tensor extension type.",
)
def test_series_filter_on_canonical_tensor_extension_array() -> None:
    arr = np.arange(20).reshape((5, 2, 2))
    data = pa.FixedShapeTensorArray.from_numpy_ndarray(arr)

    s = Series.from_arrow(data)
    assert s.datatype() == DataType.tensor(DataType.int64(), (2, 2))
    pymask = [False, True, True, None, False]
    mask = Series.from_pylist(pymask)

    result = s.filter(mask)

    assert s.datatype() == result.datatype()
    expected = [val for val, keep in zip(s.to_pylist(), pymask) if keep]
    np.testing.assert_equal(result.to_pylist(), expected)


@pytest.mark.parametrize("dtype", ARROW_INT_TYPES + ARROW_FLOAT_TYPES + ARROW_STRING_TYPES)
def test_series_filter_broadcast(dtype) -> None:
    data = pa.array([1, 2, 3, None, 5, None])

    s = Series.from_arrow(data.cast(dtype))

    mask = Series.from_pylist([False])
    result = s.filter(mask)

    assert s.datatype() == result.datatype()

    assert result.to_pylist() == []

    mask = Series.from_pylist([None]).cast(DataType.bool())
    result = s.filter(mask)
    assert s.datatype() == result.datatype()
    assert result.to_pylist() == []

    mask = Series.from_pylist([True])
    result = s.filter(mask)

    assert s.datatype() == result.datatype()

    assert result.to_pylist() == s.to_pylist()


def test_series_filter_bad_input() -> None:
    data = pa.array([1, 2, 3, None, 5, None])

    s = Series.from_arrow(data)

    mask = Series.from_pylist([False, False])
    with pytest.raises(ValueError, match="Lengths for filter do not match"):
        s.filter(mask).to_pylist()

    mask = Series.from_pylist([1])
    with pytest.raises(ValueError, match="We can only filter a Series with Boolean Series"):
        s.filter(mask).to_pylist()



from __future__ import annotations

import uuid
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from daft.dependencies import pa, pacsv, pafs, pq
from daft.filesystem import (
    _resolve_paths_and_filesystem,
    canonicalize_protocol,
    get_protocol_from_path,
)
from daft.io.delta_lake.delta_lake_write import (
    make_deltalake_add_action,
    make_deltalake_fs,
    sanitize_table_for_deltalake,
)
from daft.io.iceberg.iceberg_write import (
    coerce_pyarrow_table_to_schema,
    make_iceberg_data_file,
    make_iceberg_record,
)
from daft.recordbatch.partitioning import (
    partition_strings_to_path,
    partition_values_to_str_mapping,
)
from daft.recordbatch.recordbatch import RecordBatch
from daft.series import Series

if TYPE_CHECKING:
    from pyiceberg.schema import Schema as IcebergSchema
    from pyiceberg.table import TableProperties as IcebergTableProperties

    from daft.daft import IOConfig
    from daft.recordbatch.micropartition import MicroPartition


class FileWriterBase(ABC):
    def __init__(
        self,
        root_dir: str,
        file_idx: int,
        file_format: str,
        partition_values: RecordBatch | None = None,
        compression: str | None = None,
        io_config: IOConfig | None = None,
        version: int | None = None,
        default_partition_fallback: str | None = None,
    ):
        self.resolved_path, self.fs = self.resolve_path_and_fs(root_dir, io_config=io_config)
        self.protocol = get_protocol_from_path(root_dir)
        canonicalized_protocol = canonicalize_protocol(self.protocol)
        is_local_fs = canonicalized_protocol == "file"

        self.file_name = (
            f"{uuid.uuid4()}-{file_idx}.{file_format}"
            if version is None
            else f"{version}-{uuid.uuid4()}-{file_idx}.{file_format}"
        )
        self.partition_values = partition_values
        if self.partition_values is not None:
            self.partition_strings = {
                key: next(iter(values))
                for key, values in partition_values_to_str_mapping(self.partition_values).items()
            }
            self.dir_path = partition_strings_to_path(
                self.resolved_path,
                self.partition_strings,
                (
                    default_partition_fallback
                    if default_partition_fallback is not None
                    else "__HIVE_DEFAULT_PARTITION__"
                ),
            )
        else:
            self.partition_strings = {}
            self.dir_path = f"{self.resolved_path}"

        self.full_path = f"{self.dir_path}/{self.file_name}"
        if is_local_fs:
            self.fs.create_dir(self.dir_path, recursive=True)

        self.compression = compression if compression is not None else "none"
        self.position = 0

    def resolve_path_and_fs(self, root_dir: str, io_config: IOConfig | None = None) -> tuple[str, pafs.FileSystem]:
        [resolved_path], fs = _resolve_paths_and_filesystem(root_dir, io_config=io_config)
        return resolved_path, fs

    @abstractmethod
    def write(self, table: MicroPartition) -> int:
        """Write data to the file using the appropriate writer.

        Args:
            table: MicroPartition containing the data to be written.

        Returns:
            int: The number of bytes written to the file.
        """
        pass

    @abstractmethod
    def close(self) -> RecordBatch:
        """Close the writer and return metadata about the written file. Write should not be called after close.

        Returns:
            RecordBatch containing metadata about the written file, including path and partition values.
        """
        pass


class ParquetFileWriter(FileWriterBase):
    def __init__(
        self,
        root_dir: str,
        file_idx: int,
        partition_values: RecordBatch | None = None,
        compression: str | None = None,
        io_config: IOConfig | None = None,
        version: int | None = None,
        default_partition_fallback: str | None = None,
        metadata_collector: list[pq.FileMetaData] | None = None,
    ):
        super().__init__(
            root_dir=root_dir,
            file_idx=file_idx,
            file_format="parquet",
            partition_values=partition_values,
            compression=compression,
            io_config=io_config,
            version=version,
            default_partition_fallback=default_partition_fallback,
        )
        self.is_closed = False
        self.current_writer: pq.ParquetWriter | None = None
        self.metadata_collector: list[pq.FileMetaData] | None = metadata_collector

    def _create_writer(self, schema: pa.Schema) -> pq.ParquetWriter:
        opts = {}
        if self.metadata_collector is not None:
            opts["metadata_collector"] = self.metadata_collector
        return pq.ParquetWriter(
            self.full_path,
            schema,
            compression=self.compression,
            use_compliant_nested_type=False,
            filesystem=self.fs,
            # When using Arrow 8, it defaults to parquet version 1.
            # This hits a known bug where Arrow cannot correctly write u32 values in Parquet files:
            # https://issues.apache.org/jira/browse/ARROW-12201
            # The fix is to always use at least Parquet version 2.
            version="2.6",
            **opts,
        )

    def write(self, table: MicroPartition) -> int:
        assert not self.is_closed, "Cannot write to a closed ParquetFileWriter"
        if self.current_writer is None:
            self.current_writer = self._create_writer(table.schema().to_pyarrow_schema())
        self.current_writer.write_table(table.to_arrow(), row_group_size=len(table))

        current_position = self.current_writer.file_handle.tell()
        bytes_written = current_position - self.position
        self.position = current_position
        return bytes_written

    def close(self) -> RecordBatch:
        if self.current_writer is not None:
            self.current_writer.close()

        self.is_closed = True
        metadata = {"path": Series.from_pylist([self.full_path])}
        if self.partition_values is not None:
            for column in self.partition_values.columns():
                metadata[column.name()] = column
        return RecordBatch.from_pydict(metadata)


class CSVFileWriter(FileWriterBase):
    def __init__(
        self,
        root_dir: str,
        file_idx: int,
        partition_values: RecordBatch | None = None,
        io_config: IOConfig | None = None,
    ):
        super().__init__(
            root_dir=root_dir,
            file_idx=file_idx,
            file_format="csv",
            partition_values=partition_values,
            io_config=io_config,
        )
        self.file_handle = None
        self.current_writer: pacsv.CSVWriter | None = None
        self.is_closed = False

    def _create_writer(self, schema: pa.Schema) -> pacsv.CSVWriter:
        self.file_handle = self.fs.open_output_stream(self.full_path)
        return pacsv.CSVWriter(
            self.file_handle,
            schema,
        )

    def write(self, table: MicroPartition) -> int:
        assert not self.is_closed, "Cannot write to a closed CSVFileWriter"
        if self.current_writer is None:
            self.current_writer = self._create_writer(table.schema().to_pyarrow_schema())
        self.current_writer.write_table(table.to_arrow())

        assert self.file_handle is not None  # We should have created the file handle in _create_writer
        current_position = self.file_handle.tell()
        bytes_written = current_position - self.position
        self.position = current_position
        return bytes_written

    def close(self) -> RecordBatch:
        if self.current_writer is not None:
            self.current_writer.close()

        self.is_closed = True
        metadata = {"path": Series.from_pylist([self.full_path])}
        if self.partition_values is not None:
            for column in self.partition_values.columns():
                metadata[column.name()] = column

        return RecordBatch.from_pydict(metadata)


class IcebergWriter(ParquetFileWriter):
    def __init__(
        self,
        root_dir: str,
        file_idx: int,
        schema: IcebergSchema,
        properties: IcebergTableProperties,
        partition_spec_id: int,
        partition_values: RecordBatch | None = None,
        io_config: IOConfig | None = None,
    ):
        from pyiceberg.io.pyarrow import schema_to_pyarrow

        super().__init__(
            root_dir=root_dir,
            file_idx=file_idx,
            partition_values=partition_values,
            compression="zstd",
            io_config=io_config,
            version=None,
            default_partition_fallback="null",
            metadata_collector=[],
        )

        self.part_record = make_iceberg_record(
            partition_values.to_pylist()[0] if partition_values is not None else None
        )
        self.iceberg_schema = schema
        self.file_schema = schema_to_pyarrow(schema)
        self.partition_spec_id = partition_spec_id
        self.properties = properties

    def write(self, table: MicroPartition) -> int:
        assert not self.is_closed, "Cannot write to a closed IcebergFileWriter"
        if self.current_writer is None:
            self.current_writer = self._create_writer(self.file_schema)
        casted = coerce_pyarrow_table_to_schema(table.to_arrow(), self.file_schema)
        self.current_writer.write_table(casted)

        current_position = self.current_writer.file_handle.tell()
        bytes_written = current_position - self.position
        self.position = current_position
        return bytes_written

    def close(self) -> RecordBatch:
        if self.current_writer is not None:
            self.current_writer.close()
        self.is_closed = True

        assert self.metadata_collector is not None
        metadata = self.metadata_collector[0]
        size = self.fs.get_file_info(self.full_path).size
        path_with_protocol = f"{self.protocol}://{self.full_path}"
        data_file = make_iceberg_data_file(
            path_with_protocol,
            size,
            metadata,
            self.part_record,
            self.partition_spec_id,
            self.iceberg_schema,
            self.properties,
        )
        return RecordBatch.from_pydict({"data_file": [data_file]})


class DeltalakeWriter(ParquetFileWriter):
    def __init__(
        self,
        root_dir: str,
        file_idx: int,
        version: int,
        large_dtypes: bool,
        partition_values: RecordBatch | None = None,
        io_config: IOConfig | None = None,
    ):
        super().__init__(
            root_dir=root_dir,
            file_idx=file_idx,
            partition_values=partition_values,
            compression=None,
            io_config=io_config,
            version=version,
            default_partition_fallback=None,
            metadata_collector=[],
        )

        self.large_dtypes = large_dtypes

    def resolve_path_and_fs(self, root_dir: str, io_config: IOConfig | None = None) -> tuple[str, pafs.PyFileSystem]:
        return "", make_deltalake_fs(root_dir, io_config)

    def write(self, table: MicroPartition) -> int:
        assert not self.is_closed, "Cannot write to a closed DeltalakeFileWriter"

        converted_arrow_table = sanitize_table_for_deltalake(
            table,
            self.large_dtypes,
            self.partition_values.schema().column_names() if self.partition_values is not None else None,
        )
        if self.current_writer is None:
            self.current_writer = self._create_writer(converted_arrow_table.schema)
        self.current_writer.write_table(converted_arrow_table)

        current_position = self.current_writer.file_handle.tell()
        bytes_written = current_position - self.position
        self.position = current_position
        return bytes_written

    def close(self) -> RecordBatch:
        if self.current_writer is not None:
            self.current_writer.close()
        self.is_closed = True

        assert self.metadata_collector is not None
        metadata = self.metadata_collector[0]
        size = self.fs.get_file_info(self.full_path).size
        add_action = make_deltalake_add_action(
            path=self.full_path,
            metadata=metadata,
            size=size,
            partition_values=self.partition_strings,
        )

        return RecordBatch.from_pydict({"add_action": [add_action]})



from __future__ import annotations

import pytest

import daft
from daft import col
from daft.catalog import Catalog, Identifier, Table


def test_table_from_pydict():
    data = {"a": [1, 2, 3], "b": [4, 5, 6]}

    table = Table.from_pydict("foo", data)

    assert table.name == "foo"
    assert table.read().to_pydict() == data


def test_table_from_df():
    data = {"a": [1, 2, 3], "b": [4, 5, 6]}

    df = daft.from_pydict(data)

    table = Table.from_df("foo", df)

    assert table.name == "foo"
    assert table.read().to_pydict() == data


def test_catalog_from_pydict():
    data1 = {"a": [1, 2, 3], "b": ["a", "b", "c"]}

    data2 = {
        "c": [1, 3, 5],
    }

    data3 = {
        "d": [0.5, 1.5],
        "e": [False, True],
    }

    catalog = Catalog.from_pydict(
        {"foo": data1, "bar": daft.from_pydict(data2), "baz": Table.from_pydict("foo", data3)}, "cat"
    )

    assert catalog.name == "cat"
    assert sorted(str(ident) for ident in catalog.list_tables()) == ["bar", "baz", "foo"]
    assert catalog.has_table("foo")
    assert catalog.has_table("bar")
    assert catalog.has_table("baz")
    assert not catalog.has_table("cat")

    assert catalog.get_table("foo").read().to_pydict() == data1
    assert catalog.get_table("bar").read().to_pydict() == data2
    assert catalog.get_table("baz").read().to_pydict() == data3


def test_catalog_create_table_from_schema():
    catalog = Catalog.from_pydict({}, "cat")

    schema = daft.Schema.from_pydict(
        {
            "a": daft.DataType.int64(),
            "b": daft.DataType.string(),
        }
    )
    tbl = catalog.create_table("tab", schema)
    assert tbl.name == "tab"
    assert tbl.schema() == schema
    assert tbl.read().to_pydict() == {"a": [], "b": []}

    assert catalog.has_table("tab")
    tbl = catalog.get_table("tab")
    assert tbl.name == "tab"
    assert tbl.schema() == schema
    assert tbl.read().to_pydict() == {"a": [], "b": []}


def test_catalog_create_table_from_dataframe():
    catalog = Catalog.from_pydict({}, "cat")

    data = {"a": [1, 2, 3], "b": ["a", "b", "c"]}
    df = daft.from_pydict(data)
    tbl = catalog.create_table("tab", df)
    assert tbl.name == "tab"
    assert tbl.schema() == df.schema()
    assert tbl.read().to_pydict() == data

    assert catalog.has_table("tab")
    tbl = catalog.get_table("tab")
    assert tbl.name == "tab"
    assert tbl.schema() == df.schema()
    assert tbl.read().to_pydict() == data


def test_catalog_drop_table():
    catalog = Catalog.from_pydict({"foo": {"a": [1, 2, 3], "b": ["a", "b", "c"]}}, "cat")

    assert catalog.has_table("foo")
    catalog.drop_table("foo")
    assert not catalog.has_table("foo")

    data = {"a": [1], "b": ["x"]}

    catalog.create_table("foo", daft.from_pydict(data))
    assert catalog.has_table("foo")


def test_table_append():
    df = daft.from_pydict({"a": [1, 2, 3], "b": ["x", "y", "z"]})

    table = Table.from_df("foo", df)

    df2 = daft.from_pydict({"a": [4, 5, 6], "b": ["a", "b", "c"]})

    table.append(df2)

    assert table.read().sort(col("a")).to_pydict() == {"a": [1, 2, 3, 4, 5, 6], "b": ["x", "y", "z", "a", "b", "c"]}


def test_table_overwrite():
    df = daft.from_pydict({"a": [1, 2, 3], "b": ["x", "y", "z"]})

    table = Table.from_df("foo", df)

    df2 = daft.from_pydict({"a": [4, 5, 6], "b": ["a", "b", "c"]})

    table.overwrite(df2)

    assert table.read().sort(col("a")).to_pydict() == {"a": [4, 5, 6], "b": ["a", "b", "c"]}


def test_namespace_operations():
    catalog = Catalog.from_pydict({}, "cat")

    # Initially no namespaces
    assert catalog.list_namespaces() == []

    # Create a namespace
    catalog.create_namespace("ns1")
    assert catalog.has_namespace("ns1")
    assert catalog.list_namespaces() == [Identifier("ns1")]

    # Create a table in namespace
    data = {"a": [1, 2, 3], "b": ["x", "y", "z"]}
    catalog.create_table("ns1.table1", daft.from_pydict(data))
    assert catalog.has_table("ns1.table1")
    assert not catalog.has_table("table1")  # Table only exists in namespace

    # List tables should show namespaced table
    assert sorted(str(ident) for ident in catalog.list_tables()) == ["ns1.table1"]
    assert sorted(str(ident) for ident in catalog.list_tables("ns1")) == ["ns1.table1"]

    # Create another namespace
    catalog.create_namespace("ns2")
    assert sorted(str(ns) for ns in catalog.list_namespaces()) == ["ns1", "ns2"]

    # Drop empty namespace
    catalog.drop_namespace("ns2")
    assert not catalog.has_namespace("ns2")
    assert catalog.list_namespaces() == [Identifier("ns1")]

    # Drop namespace with content
    catalog.drop_namespace("ns1")
    assert not catalog.has_namespace("ns1")
    assert not catalog.has_table("ns1.table1")
    assert catalog.list_tables() == []


def test_namespace_errors():
    catalog = Catalog.from_pydict({}, "cat")

    # Test duplicate namespace creation
    catalog.create_namespace("ns1")
    with pytest.raises(Exception):
        catalog.create_namespace("ns1")

    # Test dropping non-existent namespace
    with pytest.raises(Exception):
        catalog.drop_namespace("nonexistent")

    # Test nested namespaces not supported
    with pytest.raises(Exception):
        catalog.create_namespace("ns1.ns2")

    # Test accessing table in non-existent namespace
    with pytest.raises(Exception):
        catalog.create_table("nonexistent.table", daft.from_pydict({"a": [1]}))


def test_namespace_table_operations():
    catalog = Catalog.from_pydict({}, "cat")

    # Create namespace and tables
    catalog.create_namespace("ns1")
    data1 = {"a": [1, 2, 3]}
    data2 = {"b": [4, 5, 6]}

    catalog.create_table("ns1.table1", daft.from_pydict(data1))
    catalog.create_table("ns1.table2", daft.from_pydict(data2))

    # Verify tables exist and have correct data
    table1 = catalog.get_table("ns1.table1")
    assert table1.read().to_pydict() == data1

    table2 = catalog.get_table("ns1.table2")
    assert table2.read().to_pydict() == data2

    # Drop table and verify namespace still exists
    catalog.drop_table("ns1.table1")
    assert not catalog.has_table("ns1.table1")
    assert catalog.has_table("ns1.table2")
    assert catalog.has_namespace("ns1")



from __future__ import annotations

import math
import os
from typing import Any

import pandas as pd
from hypothesis import note, reject, settings
from hypothesis.stateful import Bundle, RuleBasedStateMachine, precondition, rule
from hypothesis.strategies import (
    booleans,
    data,
    integers,
    lists,
    permutations,
    sampled_from,
)

import daft
from daft import DataFrame
from daft.datatype import DataType
from tests.property_based_testing.strategies import (
    columns_dict,
    generate_data,
    total_order_dtypes,
)

# Maximum number of sort columns to generate
MAX_NUM_SORT_COLS = 3


def _is_nan(obj: Any) -> bool:
    """Checks if an object is a float NaN."""
    return isinstance(obj, float) and math.isnan(obj)


@settings(max_examples=int(os.getenv("HYPOTHESIS_MAX_EXAMPLES", 100)), stateful_step_count=8, deadline=None)
class DataframeSortStateMachine(RuleBasedStateMachine):
    """Tests sorts in the face of various other operations such as filters, projections etc.

    Creates N number of sort key columns named "sort_key_{i}", and one "row_num" column which enumerates the original row number.

    Intermediate steps can consist of repartitions, filters, projections, sorts etc.

    The sort steps will additionally pull data down to check sort correctness.
    """

    Dataframes = Bundle("dataframes")

    def __init__(self):
        super().__init__()
        self.df: DataFrame | None = None
        self.sort_keys: list[str] = None
        self.num_rows_strategy = integers(min_value=8, max_value=8)
        self.repartition_num_partitions_strategy = sampled_from([1, 4, 5, 9])
        self.sorted_on: list[tuple[str, bool]] | None = None

    @rule(data=data(), num_sort_cols=integers(min_value=1, max_value=MAX_NUM_SORT_COLS))
    @precondition(lambda self: self.df is None)
    def newdataframe(self, data, num_sort_cols):
        """Start node of the state machine, creates an initial dataframe."""
        self.sort_keys = [f"sort_key_{i}" for i in range(num_sort_cols)]

        columns_dict_data = data.draw(
            columns_dict(
                generate_columns_with_type={
                    sort_key_col_name: total_order_dtypes for sort_key_col_name in self.sort_keys
                },
                num_rows_strategy=self.num_rows_strategy,
            )
        )
        df = daft.from_pydict(columns_dict_data)
        self.df = df

    @rule(data=data())
    @precondition(lambda self: self.df is not None)
    def run_sort(self, data):
        """Run a sort on the accumulated dataframe plan."""
        sort_on = data.draw(permutations(self.sort_keys))
        descending = data.draw(lists(min_size=len(self.sort_keys), max_size=len(self.sort_keys), elements=booleans()))
        self.df = self.df.sort(sort_on, desc=descending)
        self.sorted_on = list(zip(sort_on, descending))

    @rule()
    @precondition(lambda self: self.sorted_on is not None)
    def collect_after_sort(self):
        """Optionally runs after any sort step to check that sort is maintained."""
        sorted_data = self.df.to_pydict()
        sorted_on_cols = [c for c, _ in self.sorted_on]
        sorted_on_desc = [d for _, d in self.sorted_on]

        # Ensure that key column(s) are sorted correctly
        data = zip(*[sorted_data[k] for k in sorted_on_cols])

        try:
            current_tup = next(data)
        except StopIteration:
            # Trivial case, no rows to check that were sorted.
            self.sorted_on = None
            return

        try:
            for next_tup in data:
                note(f"Comparing {current_tup} and {next_tup} for desc={sorted_on_desc}")

                for current_val, next_val, desc in zip(current_tup, next_tup, sorted_on_desc):
                    a, b = (current_val, next_val) if desc else (next_val, current_val)

                    # Assert that a >= b, where the ordering is defined as: None > NaN > other values
                    # `continue` checking lex sort if values are equal, but `break` if they are not equal
                    if a is None and b is None:
                        continue
                    elif _is_nan(a) and _is_nan(b):
                        continue
                    elif a is None and _is_nan(b):
                        break
                    elif _is_nan(a) and b is None:
                        raise AssertionError(
                            f"current_val={current_val} vs next_val={next_val} is an invalid sort order for desc={desc}"
                        )
                    elif a is None or _is_nan(a):
                        break
                    elif b is None or _is_nan(b):
                        raise AssertionError(
                            f"current_val={current_val} vs next_val={next_val} is an invalid sort order for desc={desc}"
                        )

                    # Invariant here: all cases handled for None and NaN values
                    assert a is not None and not _is_nan(a)
                    assert b is not None and not _is_nan(b)

                    assert (
                        a >= b
                    ), f"current_val={current_val} vs next_val={next_val} is an invalid sort order for desc={desc}"
                    if a != b:
                        break

                current_tup = next_tup
        except AssertionError:
            sorted_keys_df = pd.DataFrame({k: pd.Series(v, dtype="object") for k, v in sorted_data.items()})
            note(f"Received sorted keys:\n{sorted_keys_df[sorted_on_cols]}")
            raise

        # Ensure that we reset self.sorted_on so that we won't try to collect again
        self.sorted_on = None

    ###
    # Intermediate fuzzing steps - these steps perform actions that should not affect the final sort result
    # Some steps are skipped because they encounter bugs that need to be fixed.
    ###

    @rule(data=data())
    @precondition(lambda self: self.df is not None)
    def repartition_df(self, data):
        """Runs a repartitioning step."""
        num_partitions = data.draw(
            self.repartition_num_partitions_strategy, label="Number of partitions for repartitioning"
        )
        self.df = self.df.repartition(num_partitions)

        # Repartitioning changes the ordering of the data, so we cannot sort after this step
        self.sorted_on = None

    @rule(data=data())
    @precondition(lambda self: self.df is not None)
    def filter_df(self, data):
        """Runs a filter on a simple equality predicate on a random column."""
        assert self.df is not None
        col_name_to_filter = data.draw(sampled_from(self.df.schema().column_names()), label="Column to filter on")
        col_daft_type = self.df.schema()[col_name_to_filter].dtype

        # Logical types do not accept equality operators, but can be filtered on by themselves
        if col_daft_type == DataType.bool():
            self.df = self.df.where(self.df[col_name_to_filter])
        # Reject if filtering on a null column - not a meaningful operation
        elif col_daft_type == DataType.null():
            reject()
        # Reject for binary types because they are not comparable yet (TODO: https://github.com/Eventual-Inc/Daft/issues/688)
        elif col_daft_type == DataType.binary():
            reject()
        else:
            filter_value = data.draw(generate_data(col_daft_type), label="Filter value")
            self.df = self.df.where(self.df[col_name_to_filter] == filter_value)

    @rule(data=data())
    @precondition(lambda self: self.df is not None)
    def project_df(self, data):
        """Runs a projection on a random column, replacing it."""
        assert self.df is not None
        column_name = data.draw(sampled_from(self.df.schema().column_names()), label="Column to filter on")
        column_daft_type = self.df.schema()[column_name].dtype
        type_to_op_mapping = {
            DataType.string(): lambda e, other: e + other,
            DataType.int64(): lambda e, other: e + other,
            DataType.int32(): lambda e, other: e + other,
            DataType.float64(): lambda e, other: e + other,
            DataType.bool(): lambda e, other: e | other,
            # No meaningful binary operations supported for these yet
            DataType.date(): lambda e, other: e.dt.year(),
            DataType.binary(): lambda e, other: e,
            DataType.null(): lambda e, other: e,
        }
        op = type_to_op_mapping[column_daft_type]
        other_binary_value = data.draw(generate_data(column_daft_type), label="Binary *other* value")
        self.df = self.df.with_column(column_name, op(self.df[column_name], other_binary_value))

        # Some of the projections change the ordering of the data, so data is no longer sorted
        self.sorted_on = None


TestDataframeSortStateMachine = DataframeSortStateMachine.TestCase
